"""Tests for OpenTelemetry sampler initialization functionality."""

from opentelemetry.sdk.trace.sampling import TraceIdRatioBased
from opentelemetry.sdk.trace.export.in_memory_span_exporter import InMemorySpanExporter
from traceloop.sdk import Traceloop
from traceloop.sdk.tracing.tracing import TracerWrapper


class TestSamplerInitialization:
    """Test class for sampler initialization functionality."""

    def test_init_with_rate_based_sampler(self):
        """Test initialization with TraceIdRatioBased sampler."""
        if hasattr(TracerWrapper, "instance"):
            _trace_wrapper_instance = TracerWrapper.instance
            del TracerWrapper.instance

        try:
            exporter = InMemorySpanExporter()
            sampler = TraceIdRatioBased(0.5)  # 50% sampling rate

            client = Traceloop.init(
                app_name="test-rate-based-sampler",
                sampler=sampler,
                exporter=exporter,
                disable_batch=True
            )

            assert client is None
            assert hasattr(TracerWrapper, "instance")
            assert TracerWrapper.instance is not None

        finally:
            if '_trace_wrapper_instance' in locals():
                TracerWrapper.instance = _trace_wrapper_instance



import pytest
import tempfile
import os


@pytest.mark.vcr
def test_create_dataset_with_columns(datasets):
    """Test creating a dataset with different column types using real API calls"""

    unique_slug = "test-columns-dataset"

    csv_content = """Name,Price,InStock,Rating
Product A,99.99,true,4.5
Product B,149.99,false,3.8
Product C,79.99,true,4.2"""

    with tempfile.NamedTemporaryFile(mode="w", suffix=".csv", delete=False) as f:
        f.write(csv_content)
        f.flush()
        csv_path = f.name

    dataset = datasets.from_csv(
        file_path=csv_path,
        slug=unique_slug,
        name="Test Columns Dataset",
        description="Dataset for testing column operations",
    )

    assert dataset is not None
    assert dataset.slug == unique_slug
    assert len(dataset.columns) >= 4  # Name, Price, InStock, Rating

    # Check that we have columns with different names
    column_names = [col.name.lower() for col in dataset.columns]
    assert any("name" in name for name in column_names)
    assert any("price" in name for name in column_names)

    os.unlink(csv_path)


@pytest.mark.vcr
def test_get_dataset_with_columns(datasets):
    """Test retrieving a dataset and checking its columns"""

    dataset = datasets.get_by_slug("test-qa")

    assert dataset is not None
    assert len(dataset.columns) >= 0  # Allow any number of columns
    assert len(dataset.rows) >= 0  # Allow any number of rows

    for column in dataset.columns:
        assert hasattr(column, "name")
        assert hasattr(column, "type")
        assert hasattr(column, "id") or hasattr(column, "slug")


@pytest.mark.vcr
def test_dataset_operations_errors(datasets):
    """Test various error conditions for dataset operations"""
    try:
        # Test with completely invalid slug
        dataset = datasets.get_by_slug("invalid-dataset-name-12345")

        assert dataset is not None

    except Exception as e:
        # Should get appropriate error for non-existent dataset
        assert "Failed to get dataset" in str(e) or "404" in str(e) or "401" in str(e)



from typing import Dict, Any, Optional, List
from pydantic import BaseModel


class EvaluatorDetails(BaseModel):
    """
    Details for configuring an evaluator.

    Args:
        slug: The evaluator slug/identifier
        version: Optional version of the evaluator
        config: Optional configuration dictionary for the evaluator
        required_input_fields: Optional list of required fields to the evaluator
            input. These fields must be present in the task output.

    Example:
        >>> EvaluatorDetails(slug="pii-detector", config={"probability_threshold": 0.8}, required_input_fields=["text"])
        >>> EvaluatorDetails(slug="my-custom-evaluator", version="v2")
    """
    slug: str
    version: Optional[str] = None
    config: Optional[Dict[str, Any]] = None
    required_input_fields: Optional[List[str]] = None



from datetime import datetime
from typing import List, Dict, Any, Optional, Union
from pydantic import BaseModel, Field
from traceloop.sdk.evaluator.config import EvaluatorDetails

EvaluatorSpec = Union[str, EvaluatorDetails]


class TaskResponse(BaseModel):
    """Model for a single task process (row)"""

    task_result: Optional[Dict[str, Any]] = None
    evaluations: Dict[str, Any] = Field(default_factory=dict)
    error: Optional[str] = None


class InitExperimentRequest(BaseModel):
    """Model for initializing an experiment"""

    slug: str
    dataset_slug: Optional[str] = None
    dataset_version: Optional[str] = None
    evaluator_slugs: Optional[List[str]] = None
    experiment_metadata: Optional[Dict[str, Any]] = None
    experiment_run_metadata: Optional[Dict[str, Any]] = None


class ExperimentResponse(BaseModel):
    """Model for experiment response"""

    id: str
    slug: str
    metadata: Optional[Dict[str, Any]] = None
    created_at: datetime
    updated_at: datetime


class ExperimentRunResponse(BaseModel):
    """Model for experiment run response"""

    id: str
    metadata: Optional[Dict[str, Any]] = None
    dataset_id: Optional[str] = None
    dataset_version: Optional[str] = None
    evaluator_ids: Optional[List[str]] = None
    created_at: datetime
    updated_at: datetime


class ExperimentInitResponse(BaseModel):
    """Model for experiment and run response"""

    experiment: ExperimentResponse
    run: ExperimentRunResponse


class CreateTaskRequest(BaseModel):
    """Model for create task request"""

    input: Dict[str, Any]
    output: Dict[str, Any]


class CreateTaskResponse(BaseModel):
    """Model for create task response"""

    id: str


class TaskResult(BaseModel):
    """Model for a single task result"""

    input: Dict[str, Any]
    output: Optional[Dict[str, Any]] = None
    error: Optional[str] = None


class GithubContext(BaseModel):
    """Model for GitHub context"""

    repository: str
    pr_url: str
    commit_hash: str
    actor: str


class RunInGithubRequest(BaseModel):
    """Model for bulk GitHub experiment execution request"""

    experiment_slug: str
    dataset_slug: Optional[str] = None
    dataset_version: Optional[str] = None
    evaluator_slugs: Optional[List[str]] = None
    task_results: List[TaskResult]
    github_context: GithubContext
    experiment_metadata: Optional[Dict[str, Any]] = None
    experiment_run_metadata: Optional[Dict[str, Any]] = None


class RunInGithubResponse(BaseModel):
    """Model for bulk GitHub experiment execution response"""

    experiment_id: str
    experiment_slug: str
    run_id: str



import pytest
import tempfile
import os


@pytest.mark.vcr
def test_create_dataset_and_add_rows(datasets):
    """Test creating a dataset and adding rows using real API calls"""

    unique_slug = "test-rows"

    csv_content = """Name,Age,Active
John,25,true
Jane,30,false
Bob,35,true"""

    with tempfile.NamedTemporaryFile(mode="w", suffix=".csv", delete=False) as f:
        f.write(csv_content)
        f.flush()
        csv_path = f.name

        dataset = datasets.from_csv(
            file_path=csv_path,
            slug=unique_slug,
            name="Test Rows Dataset",
            description="Dataset for testing row operations",
        )

        assert dataset is not None
        assert dataset.slug == unique_slug
        assert len(dataset.columns) >= 3  # Name, Age, Active
        assert len(dataset.rows) >= 0  # Allow any number of initial rows

        os.unlink(csv_path)


@pytest.mark.vcr
def test_add_rows(datasets):
    """Test the add_rows method that makes POST to /datasets/{slug}/rows"""

    unique_slug = "test-add-rows"

    # Create a simple CSV for the initial dataset
    csv_content = """Name,Age,Active
John,25,true
Jane,30,false"""

    with tempfile.NamedTemporaryFile(mode="w", suffix=".csv", delete=False) as f:
        f.write(csv_content)
        f.flush()
        csv_path = f.name

        dataset = datasets.from_csv(
            file_path=csv_path,
            slug=unique_slug,
            name="Test Add Rows Dataset",
            description="Dataset for testing add_rows method",
        )

        assert dataset is not None
        initial_row_count = len(dataset.rows) if dataset.rows else 0

        # Now test the add_rows method specifically
        new_rows = [{"name": "Alice", "age": "28", "active": "true"}]
        dataset.add_rows(new_rows)

        # Verify the row was added
        assert dataset.rows is not None
        assert len(dataset.rows) == initial_row_count + 1
        assert any(row.values["name"] == "Alice" for row in dataset.rows)

    os.unlink(csv_path)


@pytest.mark.vcr
def test_dataset_row_operations_api_errors(datasets):
    """Test handling of API errors for row operations"""
    try:
        # Try to get a non-existent dataset to record error response
        dataset = datasets.get_by_slug("definitely-non-existent-dataset-12345")
        # If we get here, the dataset unexpectedly exists, which is also valid for testing
        assert dataset is not None

    except Exception as e:
        # Should get a "dataset not found" type error
        assert "Failed to get dataset" in str(e) or "404" in str(e) or "401" in str(e)


@pytest.mark.vcr
def test_dataset_deletion(datasets):
    """Test dataset deletion using real API calls"""
    datasets.delete_by_slug("test-csv-dataset")



import json

import pytest
from openai import OpenAI
from opentelemetry.semconv._incubating.attributes import (
    gen_ai_attributes as GenAIAttributes,
)
from traceloop.sdk.prompts import get_prompt
from traceloop.sdk.prompts.client import PromptRegistryClient

prompts_json = """
{
  "prompts": [
    {
      "id": "clpuabf3h0002kdmoz3k3hesu",
      "created_at": "2023-12-06T21:31:58.637Z",
      "updated_at": "2023-12-06T21:31:58.637Z",
      "org_id": "82930720-5041-4deb-b1a5-705cc46bda3c",
      "key": "joke_generator",
      "versions": [
        {
          "id": "clpuabf780000106giwxkinpq",
          "prompt_id": "clpuabf3h0002kdmoz3k3hesu",
          "created_at": "2023-12-06T21:31:58.772Z",
          "updated_at": "2024-01-07T08:25:17.194Z",
          "author": "tomer f",
          "org_id": "82930720-5041-4deb-b1a5-705cc46bda3c",
          "hash": "8296b2fe",
          "version": 0,
          "name": "dfgdfg",
          "provider": "openai",
          "templating_engine": "jinja2",
          "messages": [
            {
              "role": "user",
              "index": 0,
              "template": "Tell me a joke about OpenTelemetry, {{style}} style",
              "variables": [
                "style"
              ]
            }
          ],
          "llm_config": {
            "mode": "chat",
            "stop": [],
            "model": "gpt-3.5-turbo",
            "top_p": 1,
            "temperature": 1,
            "presence_penalty": 0,
            "frequency_penalty": 0
          },
          "publishable": true
        }
      ],
      "target": {
        "id": "clr38ainb0001mrby3q81lec6",
        "updated_at": "2024-01-07T08:24:55.271Z",
        "prompt_id": "clpuabf3h0002kdmoz3k3hesu",
        "org_id": "82930720-5041-4deb-b1a5-705cc46bda3c",
        "environment": "dev",
        "version": "clpuabf780000106giwxkinpq"
      }
    }
  ],
  "environment": "dev"
}
"""

prompts_with_tools_json = """
{
  "prompts": [
    {
      "id": "clpuabf3h0002kdmoz3k3hesu",
      "created_at": "2023-12-06T21:31:58.637Z",
      "updated_at": "2023-12-06T21:31:58.637Z",
      "org_id": "82930720-5041-4deb-b1a5-705cc46bda3c",
      "key": "joke_generator",
      "versions": [
        {
          "id": "clpuabf780000106giwxkinpq",
          "prompt_id": "clpuabf3h0002kdmoz3k3hesu",
          "created_at": "2023-12-06T21:31:58.772Z",
          "updated_at": "2024-01-07T08:25:17.194Z",
          "author": "tomer f",
          "org_id": "82930720-5041-4deb-b1a5-705cc46bda3c",
          "hash": "8296b2fe",
          "version": 0,
          "name": "dfgdfg",
          "provider": "openai",
          "templating_engine": "jinja2",
          "messages": [
            {
              "role": "user",
              "index": 0,
              "template": "Tell me a joke about OpenTelemetry, {{style}} style",
              "variables": [
                "style"
              ]
            }
          ],
          "llm_config": {
            "mode": "chat",
            "stop": [],
            "model": "gpt-3.5-turbo",
            "top_p": 1,
            "temperature": 1,
            "presence_penalty": 0,
            "frequency_penalty": 0,
            "tool_choice": "required",
            "tools": [
              {
                "type": "function",
                "function": {
                  "name": "get_joke",
                  "description": "Get a joke about OpenTelemetry",
                  "parameters": {
                    "type": "object",
                    "properties": {
                      "style": {
                        "type": "string",
                        "description": "The style of the joke"
                      }
                    },
                    "required": ["style"]
                  }
                }
              }
            ]
          },
          "publishable": true
        }
      ],
      "target": {
        "id": "clr38ainb0001mrby3q81lec6",
        "updated_at": "2024-01-07T08:24:55.271Z",
        "prompt_id": "clpuabf3h0002kdmoz3k3hesu",
        "org_id": "82930720-5041-4deb-b1a5-705cc46bda3c",
        "environment": "dev",
        "version": "clpuabf780000106giwxkinpq"
      }
    }
  ],
  "environment": "dev"
}
"""

prompts_with_response_format_json = """

{
  "prompts": [
    {
      "id": "clpuabf3h0002kdmoz3k3hesu",
      "created_at": "2023-12-06T21:31:58.637Z",
      "updated_at": "2023-12-06T21:31:58.637Z",
      "org_id": "82930720-5041-4deb-b1a5-705cc46bda3c",
      "key": "joke_generator",
      "versions": [
        {
          "id": "clpuabf780000106giwxkinpq",
          "prompt_id": "clpuabf3h0002kdmoz3k3hesu",
          "created_at": "2023-12-06T21:31:58.772Z",
          "updated_at": "2024-01-07T08:25:17.194Z",
          "author": "tomer f",
          "org_id": "82930720-5041-4deb-b1a5-705cc46bda3c",
          "hash": "8296b2fe",
          "version": 0,
          "name": "dfgdfg",
          "provider": "openai",
          "templating_engine": "jinja2",
          "messages": [
            {
              "role": "user",
              "index": 0,
              "template": "Tell me a joke about OpenTelemetry, {{style}} style",
              "variables": [
                "style"
              ]
            }
          ],
          "llm_config": {
            "mode": "chat",
            "stop": [],
            "model": "gpt-4o-mini",
            "top_p": 1,
            "presence_penalty": 0,
            "frequency_penalty": 0,
            "temperature": 1,
            "response_format": {
              "type": "json_schema",
              "json_schema": {
                "name": "joke",
                "schema": {
                  "type": "object",
                  "properties": {
                    "joke": {
                      "type": "string",
                      "description": "The joke"
                    }
                  }
                }
              }
            }
          },
          "publishable": true
        }
      ],
      "target": {
        "id": "clr38ainb0001mrby3q81lec6",
        "updated_at": "2024-01-07T08:24:55.271Z",
        "prompt_id": "clpuabf3h0002kdmoz3k3hesu",
        "org_id": "82930720-5041-4deb-b1a5-705cc46bda3c",
        "environment": "dev",
        "version": "clpuabf780000106giwxkinpq"
      }
    }
  ]
}
"""


@pytest.fixture
def openai_client():
    return OpenAI()


@pytest.mark.vcr
def test_prompt_management(exporter, openai_client):
    PromptRegistryClient()._registry.load(prompts_json=json.loads(prompts_json))
    prompt_args = get_prompt(key="joke_generator", variables={"style": "pirate"})
    openai_client.chat.completions.create(**prompt_args)

    spans = exporter.get_finished_spans()
    assert [span.name for span in spans] == [
        "openai.chat",
    ]
    open_ai_span = spans[0]
    assert (
        open_ai_span.attributes[f"{GenAIAttributes.GEN_AI_PROMPT}.0.content"]
        == "Tell me a joke about OpenTelemetry, pirate style"
    )
    assert open_ai_span.attributes.get(f"{GenAIAttributes.GEN_AI_COMPLETION}.0.content")
    assert open_ai_span.attributes.get("traceloop.prompt.key") == "joke_generator"


@pytest.mark.vcr
def test_prompt_management_with_tools(exporter, openai_client):
    PromptRegistryClient()._registry.load(
        prompts_json=json.loads(prompts_with_tools_json)
    )
    prompt_args = get_prompt(key="joke_generator", variables={"style": "pirate"})
    openai_client.chat.completions.create(**prompt_args)

    spans = exporter.get_finished_spans()
    open_ai_span = spans[0]
    completion = open_ai_span.attributes.get(
        f"{GenAIAttributes.GEN_AI_COMPLETION}.0.tool_calls.0.name"
    )
    assert completion == "get_joke"


@pytest.mark.vcr
def test_prompt_management_with_response_format(exporter, openai_client):
    PromptRegistryClient()._registry.load(
        prompts_json=json.loads(prompts_with_response_format_json)
    )
    prompt_args = get_prompt(key="joke_generator", variables={"style": "pirate"})
    openai_client.chat.completions.create(**prompt_args)

    spans = exporter.get_finished_spans()
    open_ai_span = spans[0]
    completion = open_ai_span.attributes.get(
        f"{GenAIAttributes.GEN_AI_COMPLETION}.0.content"
    )
    try:
        json.loads(completion)
    except json.JSONDecodeError:
        pytest.fail("Response is not valid JSON")
    assert True



# generated by datamodel-codegen:
#   filename:  tmpslpb59ih.json

from __future__ import annotations

from pydantic import BaseModel, Field


class AgentEfficiencyResponse(BaseModel):
    step_efficiency_reason: str | None = Field(
        None, examples=['Agent completed task with minimal redundant steps']
    )
    step_efficiency_score: float | None = Field(None, examples=[0.85])
    task_completion_reason: str | None = Field(
        None, examples=['All required tasks were completed successfully']
    )
    task_completion_score: float | None = Field(None, examples=[0.92])


class AgentFlowQualityResponse(BaseModel):
    reason: str | None = Field(
        None, examples=['Agent followed the expected flow correctly']
    )
    score: float | None = Field(None, examples=[0.89])
    success: bool | None = Field(None, examples=[True])


class AgentGoalAccuracyResponse(BaseModel):
    accuracy_score: float | None = Field(None, examples=[0.88])


class AgentGoalCompletenessResponse(BaseModel):
    reason: str | None = Field(None, examples=['All user goals were accomplished'])
    score: float | None = Field(None, examples=[0.95])
    success: bool | None = Field(None, examples=[True])


class AgentToolErrorDetectorResponse(BaseModel):
    reason: str | None = Field(
        None, examples=['Tool executed successfully without errors']
    )
    success: bool | None = Field(None, examples=[True])


class AgentToolTrajectoryResponse(BaseModel):
    reason: str | None = Field(
        None, examples=['Tool calls match the expected trajectory']
    )
    score: float | None = Field(None, examples=[0.85])
    success: bool | None = Field(None, examples=[True])


class AnswerCompletenessResponse(BaseModel):
    answer_completeness_score: float | None = Field(None, examples=[0.95])


class AnswerCorrectnessResponse(BaseModel):
    correctness_score: float | None = Field(None, examples=[0.91])


class AnswerRelevancyResponse(BaseModel):
    is_relevant: bool | None = Field(None, examples=[True])


class CharCountRatioResponse(BaseModel):
    char_ratio: float | None = Field(None, examples=[0.75])


class CharCountResponse(BaseModel):
    char_count: int | None = Field(None, examples=[42])


class ContextRelevanceResponse(BaseModel):
    relevance_score: float | None = Field(None, examples=[0.88])


class ConversationQualityResponse(BaseModel):
    conversation_quality_score: float | None = Field(None, examples=[0.82])


class ErrorResponse(BaseModel):
    error: str | None = Field(None, examples=['error message'])


class FaithfulnessResponse(BaseModel):
    is_faithful: bool | None = Field(None, examples=[True])


class HtmlComparisonResponse(BaseModel):
    similarity_score: float | None = Field(None, examples=[0.92])


class InstructionAdherenceResponse(BaseModel):
    instruction_adherence_score: float | None = Field(None, examples=[0.87])


class IntentChangeResponse(BaseModel):
    reason: str | None = Field(
        None, examples=['User intent remained consistent throughout the conversation']
    )
    score: int | None = Field(None, examples=[1])
    success: bool | None = Field(None, examples=[True])


class JSONValidatorResponse(BaseModel):
    is_valid_json: bool | None = Field(None, examples=[True])


class PIIDetectorResponse(BaseModel):
    has_pii: bool | None = Field(None, examples=[False])


class PerplexityResponse(BaseModel):
    perplexity_score: float | None = Field(None, examples=[12.5])


class PlaceholderRegexResponse(BaseModel):
    is_valid_regex: bool | None = Field(None, examples=[True])


class ProfanityDetectorResponse(BaseModel):
    is_safe: bool | None = Field(None, examples=[False])


class PromptInjectionResponse(BaseModel):
    has_injection: bool | None = Field(None, examples=[False])


class PromptPerplexityResponse(BaseModel):
    perplexity_score: float | None = Field(None, examples=[8.3])


class RegexValidatorResponse(PlaceholderRegexResponse):
    pass


class SQLValidatorResponse(BaseModel):
    is_valid_sql: bool | None = Field(None, examples=[True])


class SecretsDetectorResponse(BaseModel):
    has_secret: bool | None = Field(None, examples=[False])


class SemanticSimilarityResponse(HtmlComparisonResponse):
    pass


class SexismDetectorResponse(ProfanityDetectorResponse):
    pass


class ToneDetectionResponse(BaseModel):
    score: float | None = Field(None, examples=[0.95])
    tone: str | None = Field(None, examples=['neutral'])


class TopicAdherenceResponse(BaseModel):
    adherence_score: float | None = Field(None, examples=[0.95])


class ToxicityDetectorResponse(ProfanityDetectorResponse):
    pass


class UncertaintyDetectorResponse(BaseModel):
    answer: str | None = Field(None, examples=['Paris'])
    uncertainty: float | None = Field(None, examples=[0.95])


class WordCountRatioResponse(BaseModel):
    word_ratio: float | None = Field(None, examples=[0.85])


class WordCountResponse(BaseModel):
    word_count: int | None = Field(None, examples=[10])



from typing import Any, Optional, TypeVar, Callable
import warnings
import inspect
from functools import wraps

from opentelemetry.semconv_ai import TraceloopSpanKindValues

from traceloop.sdk.decorators.base import (
    entity_class,
    entity_method,
)

F = TypeVar("F", bound=Callable[..., Any])

__all__ = [
    "conversation",
    "task",
    "workflow",
    "agent",
    "tool",
    "atask",
    "aworkflow",
    "aagent",
    "atool",
]


def conversation(conversation_id: str) -> Callable[[F], F]:
    """
    Decorator to set the conversation ID for all spans within the decorated function.

    Args:
        conversation_id: Unique identifier for the conversation

    Example:
        @conversation(conversation_id="conv-123")
        def handle_chat(user_message: str):
            response = llm.chat(user_message)
            return response

        # Can be combined with @workflow:
        @workflow(name="chat_session")
        @conversation(conversation_id="conv-456")
        def handle_chat_with_workflow(user_message: str):
            response = llm.chat(user_message)
            return response
    """
    from traceloop.sdk.tracing.tracing import set_conversation_id

    def decorator(fn: F) -> F:
        if inspect.iscoroutinefunction(fn):
            @wraps(fn)
            async def async_wrapper(*args, **kwargs):
                set_conversation_id(conversation_id)
                return await fn(*args, **kwargs)
            return async_wrapper
        else:
            @wraps(fn)
            def sync_wrapper(*args, **kwargs):
                set_conversation_id(conversation_id)
                return fn(*args, **kwargs)
            return sync_wrapper

    return decorator


def task(
    name: Optional[str] = None,
    version: Optional[int] = None,
    method_name: Optional[str] = None,
    tlp_span_kind: Optional[TraceloopSpanKindValues] = TraceloopSpanKindValues.TASK,
) -> Callable[[F], F]:
    if method_name is None:
        return entity_method(name=name, version=version, tlp_span_kind=tlp_span_kind)
    else:
        return entity_class(
            name=name,
            version=version,
            method_name=method_name,
            tlp_span_kind=tlp_span_kind,
        )


def workflow(
    name: Optional[str] = None,
    version: Optional[int] = None,
    method_name: Optional[str] = None,
    tlp_span_kind: Optional[TraceloopSpanKindValues] = TraceloopSpanKindValues.WORKFLOW,
) -> Callable[[F], F]:
    if method_name is None:
        return entity_method(name=name, version=version, tlp_span_kind=tlp_span_kind)
    else:
        return entity_class(
            name=name,
            version=version,
            method_name=method_name,
            tlp_span_kind=tlp_span_kind,
        )


def agent(
    name: Optional[str] = None,
    version: Optional[int] = None,
    method_name: Optional[str] = None,
) -> Callable[[F], F]:
    return workflow(
        name=name,
        version=version,
        method_name=method_name,
        tlp_span_kind=TraceloopSpanKindValues.AGENT,
    )


def tool(
    name: Optional[str] = None,
    version: Optional[int] = None,
    method_name: Optional[str] = None,
) -> Callable[[F], F]:
    return task(
        name=name,
        version=version,
        method_name=method_name,
        tlp_span_kind=TraceloopSpanKindValues.TOOL,
    )


# Async Decorators - Deprecated
def atask(
    name: Optional[str] = None,
    version: Optional[int] = None,
    method_name: Optional[str] = None,
    tlp_span_kind: Optional[TraceloopSpanKindValues] = TraceloopSpanKindValues.TASK,
) -> Callable[[F], F]:
    warnings.warn(
        "DeprecationWarning: The @atask decorator will be removed in a future version. "
        "Please migrate to @task for both sync and async operations.",
        DeprecationWarning,
        stacklevel=2,
    )
    if method_name is None:
        return entity_method(name=name, version=version, tlp_span_kind=tlp_span_kind)
    else:
        return entity_class(
            name=name,
            version=version,
            method_name=method_name,
            tlp_span_kind=tlp_span_kind,
        )


def aworkflow(
    name: Optional[str] = None,
    version: Optional[int] = None,
    method_name: Optional[str] = None,
    tlp_span_kind: Optional[TraceloopSpanKindValues] = TraceloopSpanKindValues.WORKFLOW,
) -> Callable[[F], F]:
    warnings.warn(
        "DeprecationWarning: The @aworkflow decorator will be removed in a future version. "
        "Please migrate to @workflow for both sync and async operations.",
        DeprecationWarning,
        stacklevel=2,
    )
    if method_name is None:
        return entity_method(name=name, version=version, tlp_span_kind=tlp_span_kind)
    else:
        return entity_class(
            name=name,
            version=version,
            method_name=method_name,
            tlp_span_kind=tlp_span_kind,
        )


def aagent(
    name: Optional[str] = None,
    version: Optional[int] = None,
    method_name: Optional[str] = None,
) -> Callable[[F], F]:
    warnings.warn(
        "DeprecationWarning: The @aagent decorator will be removed in a future version. "
        "Please migrate to @agent for both sync and async operations.",
        DeprecationWarning,
        stacklevel=2,
    )
    return atask(
        name=name,
        version=version,
        method_name=method_name,
        tlp_span_kind=TraceloopSpanKindValues.AGENT,
    )


def atool(
    name: Optional[str] = None,
    version: Optional[int] = None,
    method_name: Optional[str] = None,
) -> Callable[[F], F]:
    warnings.warn(
        "DeprecationWarning: The @atool decorator will be removed in a future version. "
        "Please migrate to @tool for both sync and async operations.",
        DeprecationWarning,
        stacklevel=2,
    )
    return atask(
        name=name,
        version=version,
        method_name=method_name,
        tlp_span_kind=TraceloopSpanKindValues.TOOL,
    )



import pytest
from traceloop.sdk.datasets.dataset import Dataset
from traceloop.sdk.datasets.model import DatasetMetadata


@pytest.mark.vcr
def test_get_dataset_by_slug(datasets):
    dataset = datasets.get_by_slug("test-qa")

    assert isinstance(dataset, Dataset)
    assert dataset.slug == "test-qa"
    assert hasattr(dataset, "name")
    assert hasattr(dataset, "description")


@pytest.mark.vcr
def test_get_all_datasets(datasets):
    datasets_list = datasets.get_all()

    assert isinstance(datasets_list, list)
    assert len(datasets_list) >= 0

    for dataset in datasets_list:
        assert isinstance(dataset, DatasetMetadata)
        assert hasattr(dataset, "id")
        assert hasattr(dataset, "slug")
        assert hasattr(dataset, "name")


@pytest.mark.vcr
def test_get_version_csv(datasets):
    try:
        csv_data = datasets.get_version_csv(slug="test-qa", version="v1")
        assert isinstance(csv_data, str)
    except Exception as e:
        # Allow for expected API errors during recording (dataset might not exist)
        assert "Failed to get dataset" in str(e) or "404" in str(e) or "401" in str(e)


@pytest.mark.vcr
def test_delete_by_slug(datasets):
    try:
        # Use a test dataset that's safe to delete
        datasets.delete_by_slug("test-csv-dataset-conflict")
    except Exception as e:
        # Allow for expected API errors (dataset might not exist)
        assert (
            "Failed to delete dataset" in str(e) or "404" in str(e) or "401" in str(e)
        )


@pytest.mark.vcr
def test_delete_by_slug_failure(datasets):
    with pytest.raises(Exception) as exc_info:
        datasets.delete_by_slug("non-existent-dataset-123")

    # The exact error message may vary based on the recorded API response
    assert "Failed to delete dataset" in str(exc_info.value) or "404" in str(
        exc_info.value
    )


@pytest.mark.vcr
def test_get_all_datasets_with_invalid_credentials():
    # Test with invalid API key to record failure case
    from traceloop.sdk.client.http import HTTPClient
    from traceloop.sdk.datasets.datasets import Datasets

    http = HTTPClient(
        base_url="https://api-staging.traceloop.com",
        api_key="invalid-key",
        version="1.0.0",
    )
    invalid_datasets = Datasets(http)

    try:
        invalid_datasets.get_all()
        # If this doesn't raise an exception, the test setup might be wrong
        assert False, "Expected authentication error"
    except Exception as exc_info:
        # Should get authentication error or a generic failure error when using VCR
        assert (
            "401" in str(exc_info)
            or "authentication" in str(exc_info).lower()
            or "Failed to get datasets" in str(exc_info)
        )


@pytest.mark.vcr
def test_get_dataset_by_slug_failure(datasets):
    with pytest.raises(Exception) as exc_info:
        datasets.get_by_slug("definitely-non-existent-dataset-123")

    assert "Failed to get dataset" in str(exc_info.value) or "404" in str(
        exc_info.value
    )


@pytest.mark.vcr
def test_get_version_csv_failure(datasets):
    with pytest.raises(Exception) as exc_info:
        datasets.get_version_csv("definitely-non-existent-dataset-123", "v1")

    assert "Failed to get dataset" in str(exc_info.value) or "404" in str(
        exc_info.value
    )



"""
Test constants to reduce string repetition across dataset tests.
"""


class TestConstants:
    # API Configuration
    API_KEY = "test-api-key"

    # Dataset IDs and Identifiers
    DATASET_ID = "cmdvei5dd000g01vvyftz2zv1"
    DATASET_SLUG = "test-dataset"
    DATASET_NAME = "Dataset"
    DATASET_DESCRIPTION = "Dataset Description"

    # Test Dataset Slugs
    TEST_CSV_DATASET_SLUG = "test-csv-dataset"
    TEST_DF_DATASET_SLUG = "test-df-dataset"
    PRODUCT_INVENTORY_SLUG = "product-inventory-2"

    # Test Dataset Names
    TEST_CSV_DATASET_NAME = "Test CSV Dataset"
    TEST_DF_DATASET_NAME = "Test DataFrame Dataset"
    PRODUCT_INVENTORY_NAME = "Product Inventory"

    # Test Dataset Descriptions
    TEST_CSV_DATASET_DESC = "Dataset created from CSV"
    TEST_DF_DATASET_DESC = "Dataset created from DataFrame"
    PRODUCT_INVENTORY_DESC = "Sample product inventory data"

    # Column IDs
    NAME_COLUMN_ID = "cmdvei5dd000d01vv2yvmp7vt"
    PRICE_COLUMN_ID = "cmdvei5dd000e01vv8h7k3q2s"
    STOCK_COLUMN_ID = "cmdvei5dd000f01vvmn9x1p4w"

    # Column Names
    NAME_COLUMN = "Name"
    PRICE_COLUMN = "Price"
    STOCK_COLUMN = "In Stock"
    PRODUCT_COLUMN = "product"

    # Column Types
    STRING_TYPE = "string"
    NUMBER_TYPE = "number"
    BOOLEAN_TYPE = "boolean"

    # Test Data Values
    LAPTOP_VALUE = "Laptop"
    LAPTOP_PRICE = 999.99
    MOUSE_VALUE = "Mouse"
    MOUSE_PRICE = 29.99

    # CSV Test Content
    SAMPLE_CSV_CONTENT = """Name,Price,In Stock
Laptop,999.99,true
Mouse,29.99,false"""

    # Mock Dataset IDs for different tests
    MOCK_DATASET_ID = "mock-dataset-id"
    TEST_DATASET_ID = "test_dataset_id"

    # API Endpoints
    DEFAULT_PROJECT_PATH = "datasets"

    # Version Information
    VERSION_V1 = "v1"

    # Test Row Data Column IDs (from mock_objects.py)
    TEST_NAME_COL_ID = "cmdr3ce1s0003hmp0vqons5ey"
    TEST_VALUE_COL_ID = "cmdr3ce1s0004hmp0ies575jr"
    TEST_ACTIVE_COL_ID = "cmdr3ce1s0005hmp0bdln01js"

    # File Paths
    NON_EXISTENT_FILE_PATH = "/non_existent/file.csv"



from typing import Dict, Optional, Any
from pydantic import BaseModel


class InputExtractor(BaseModel):
    source: str  # "input" or "output"
    key: Optional[str] = None  # Key to extract from
    use_regex: bool = False  # Whether to use regex pattern
    regex_pattern: Optional[str] = None  # Regex pattern to apply


class ExecuteEvaluatorRequest(BaseModel):
    input_schema_mapping: Dict[str, InputExtractor]
    evaluator_version: Optional[str] = None
    evaluator_config: Optional[Dict[str, Any]] = None


class OutputSchema(BaseModel):
    reason: Optional[str] = None
    success: bool



import pytest
from opentelemetry.semconv._incubating.attributes.gen_ai_attributes import (
    GEN_AI_CONVERSATION_ID,
)
from opentelemetry.semconv_ai import SpanAttributes
from traceloop.sdk.tracing import set_conversation_id
from traceloop.sdk.decorators import conversation, task, workflow


def test_set_conversation_id_api(exporter):
    """Test set_conversation_id() API"""

    @workflow(name="chat_workflow")
    def chat_workflow(conv_id: str):
        set_conversation_id(conv_id)
        return test_task()

    @task(name="test_task")
    def test_task():
        return "response"

    chat_workflow("conv-abc")

    spans = exporter.get_finished_spans()
    assert len(spans) == 2

    # The workflow span is created before set_conversation_id is called,
    # so it won't have conversation_id. But the task span should have it.
    task_span = next(span for span in spans if "task" in span.name)
    assert task_span.attributes[GEN_AI_CONVERSATION_ID] == "conv-abc"


def test_conversation_decorator_sync(exporter):
    """Test conversation decorator with sync function"""

    @conversation(conversation_id="conv-123")
    @workflow(name="test_conversation")
    def chat_session(message: str):
        return inner_task(message)

    @task(name="inner_task")
    def inner_task(message: str):
        return f"Response to: {message}"

    chat_session("Hello")

    spans = exporter.get_finished_spans()
    assert len(spans) == 2

    # All spans should have conversation_id
    for span in spans:
        assert span.attributes[GEN_AI_CONVERSATION_ID] == "conv-123"

    # Workflow span should have correct name
    workflow_span = next(span for span in spans if "workflow" in span.name)
    assert workflow_span.attributes[SpanAttributes.TRACELOOP_WORKFLOW_NAME] == "test_conversation"


@pytest.mark.asyncio
async def test_conversation_decorator_async(exporter):
    """Test conversation decorator with async function"""

    @conversation(conversation_id="conv-456")
    @workflow(name="async_conversation")
    async def async_chat_session(message: str):
        return await async_inner_task(message)

    @task(name="async_inner_task")
    async def async_inner_task(message: str):
        return f"Response to: {message}"

    await async_chat_session("Hello")

    spans = exporter.get_finished_spans()

    # All spans should have conversation_id
    for span in spans:
        assert span.attributes[GEN_AI_CONVERSATION_ID] == "conv-456"


def test_conversation_id_in_nested_tasks(exporter):
    """Test that conversation_id propagates through nested tasks"""

    @workflow(name="chat_workflow")
    def chat_workflow():
        set_conversation_id("conv-nested-789")
        return level1_task()

    @task(name="level1_task")
    def level1_task():
        return level2_task()

    @task(name="level2_task")
    def level2_task():
        return "deep response"

    chat_workflow()

    spans = exporter.get_finished_spans()
    assert len(spans) == 3

    # Task spans should have conversation_id (workflow span won't because it's created before set_conversation_id)
    task_spans = [span for span in spans if "task" in span.name]
    assert len(task_spans) == 2
    for span in task_spans:
        assert span.attributes[GEN_AI_CONVERSATION_ID] == "conv-nested-789"


def test_conversation_id_with_multiple_calls(exporter):
    """Test that conversation_id works correctly with multiple separate calls"""

    @workflow(name="chat_workflow")
    def chat_workflow(conv_id: str):
        set_conversation_id(conv_id)
        return test_task()

    @task(name="test_task")
    def test_task():
        return "response"

    # First conversation
    chat_workflow("conv-first")
    spans_first = exporter.get_finished_spans()
    exporter.clear()

    # Second conversation
    chat_workflow("conv-second")
    spans_second = exporter.get_finished_spans()

    # Verify first conversation task spans have correct conversation_id
    task_span_first = next(span for span in spans_first if "task" in span.name)
    assert task_span_first.attributes[GEN_AI_CONVERSATION_ID] == "conv-first"

    # Verify second conversation task spans have correct conversation_id
    task_span_second = next(span for span in spans_second if "task" in span.name)
    assert task_span_second.attributes[GEN_AI_CONVERSATION_ID] == "conv-second"


def test_conversation_decorator_standalone(exporter):
    """Test conversation decorator without workflow"""

    @conversation(conversation_id="conv-standalone")
    @task(name="chat_task")
    def my_chat_function():
        return "response"

    my_chat_function()

    spans = exporter.get_finished_spans()
    assert len(spans) == 1

    # Should have conversation_id
    assert spans[0].attributes[GEN_AI_CONVERSATION_ID] == "conv-standalone"



"""Tests for creating datasets with initial attachments."""

import os
import tempfile
from unittest.mock import patch

import pytest
from traceloop.sdk.datasets import (
    Attachment,
    ExternalAttachment,
    FileCellType,
)
from traceloop.sdk.datasets.model import (
    ColumnDefinition,
    ColumnType,
    CreateDatasetRequest,
)


@pytest.mark.vcr
def test_create_dataset_with_external_attachments(datasets):
    """Test creating dataset with ExternalAttachment objects in rows."""

    # Create dataset request with external attachments
    dataset_request = CreateDatasetRequest(
        slug="test-dataset-external-attachments",
        name="Products with External Media",
        description="Test dataset with external URLs",
        columns=[
            ColumnDefinition(slug="name", name="Product Name", type=ColumnType.STRING),
            ColumnDefinition(slug="price", name="Price", type=ColumnType.NUMBER),
            ColumnDefinition(slug="video", name="Demo Video", type=ColumnType.FILE),
            ColumnDefinition(slug="manual", name="Manual", type=ColumnType.FILE),
        ],
        rows=[
            {
                "name": "Widget Pro",
                "price": 99.99,
                "video": ExternalAttachment(
                    url="https://www.youtube.com/watch?v=demo1",
                    file_type=FileCellType.VIDEO,
                    metadata={"title": "Widget Pro Demo", "duration": "3:45"},
                ),
                "manual": ExternalAttachment(
                    url="https://docs.google.com/document/d/widget-manual",
                    file_type=FileCellType.FILE,
                    metadata={"pages": 25},
                ),
            },
            {
                "name": "Gadget Plus",
                "price": 149.99,
                "video": ExternalAttachment(
                    url="https://vimeo.com/demo2", file_type=FileCellType.VIDEO
                ),
                "manual": None,  # No manual for this product
            },
        ],
    )

    # Create the dataset
    dataset = datasets.create(dataset_request)

    # Verify dataset was created
    assert dataset.slug == "test-dataset-external-attachments"
    assert len(dataset.rows) == 2

    # Verify first row attachments
    row1 = dataset.rows[0]
    assert row1.values["name"] == "Widget Pro"
    assert row1.values["price"] == 99.99

    # Check video attachment
    video_cell = row1.values.get("video")
    assert video_cell is not None
    assert video_cell["storage"] == "external"
    assert video_cell["url"] == "https://www.youtube.com/watch?v=demo1"
    assert video_cell["type"] == "video"
    assert video_cell["status"] == "success"

    # Check manual attachment
    manual_cell = row1.values.get("manual")
    assert manual_cell is not None
    assert manual_cell["storage"] == "external"
    assert manual_cell["url"] == "https://docs.google.com/document/d/widget-manual"

    # Verify second row
    row2 = dataset.rows[1]
    assert row2.values["name"] == "Gadget Plus"
    video_cell2 = row2.values.get("video")
    assert video_cell2 is not None
    assert video_cell2["url"] == "https://vimeo.com/demo2"
    assert row2.values.get("manual") is None  # Should remain None


@pytest.mark.vcr
def test_create_dataset_with_file_attachments_mocked(datasets):
    """Test creating dataset with Attachment objects (file uploads) using mocked S3."""

    # Create test files
    test_image = tempfile.NamedTemporaryFile(suffix=".jpg", delete=False)
    test_image.write(b"fake image data")
    test_image.close()

    test_pdf = tempfile.NamedTemporaryFile(suffix=".pdf", delete=False)
    test_pdf.write(b"fake pdf data")
    test_pdf.close()

    # Create dataset request with file attachments
    dataset_request = CreateDatasetRequest(
        slug="test-dataset-file-attachments",
        name="Products with Files",
        columns=[
            ColumnDefinition(slug="name", name="Name", type=ColumnType.STRING),
            ColumnDefinition(slug="image", name="Image", type=ColumnType.FILE),
            ColumnDefinition(slug="manual", name="Manual", type=ColumnType.FILE),
        ],
        rows=[
            {
                "name": "Product A",
                "image": Attachment(
                    file_path=test_image.name,
                    file_type=FileCellType.IMAGE,
                    metadata={"alt_text": "Product A image"},
                ),
                "manual": Attachment(
                    file_path=test_pdf.name,
                    file_type=FileCellType.FILE,
                    metadata={"version": "1.0"},
                ),
            }
        ],
    )

    # Mock the S3 upload to succeed
    with patch.object(Attachment, "_upload_to_s3", return_value=True):
        dataset = datasets.create(dataset_request)

    # Verify dataset was created
    assert dataset.slug == "test-dataset-file-attachments"
    assert len(dataset.rows) == 1

    # Verify attachments were processed
    row = dataset.rows[0]
    assert row.values["name"] == "Product A"

    image_cell = row.values.get("image")
    assert image_cell is not None
    assert image_cell["storage"] == "internal"
    assert image_cell["type"] == "image"
    assert image_cell["status"] == "success"
    assert "storage_key" in image_cell

    manual_cell = row.values.get("manual")
    assert manual_cell is not None
    assert manual_cell["storage"] == "internal"
    assert manual_cell["type"] == "file"
    assert manual_cell["status"] == "success"

    # Clean up
    os.unlink(test_image.name)
    os.unlink(test_pdf.name)


@pytest.mark.vcr
def test_create_dataset_with_mixed_attachments(datasets):
    """Test creating dataset with both Attachment and ExternalAttachment objects."""

    # Create a test file
    test_file = tempfile.NamedTemporaryFile(suffix=".txt", delete=False)
    test_file.write(b"test content")
    test_file.close()

    dataset_request = CreateDatasetRequest(
        slug="test-dataset-mixed-attachments",
        name="Mixed Attachments Dataset",
        columns=[
            ColumnDefinition(slug="id", name="ID", type=ColumnType.STRING),
            ColumnDefinition(slug="file", name="File", type=ColumnType.FILE),
            ColumnDefinition(slug="video", name="Video", type=ColumnType.FILE),
        ],
        rows=[
            {
                "id": "001",
                "file": Attachment(
                    file_path=test_file.name, file_type=FileCellType.FILE
                ),
                "video": ExternalAttachment(
                    url="https://youtube.com/watch?v=test", file_type=FileCellType.VIDEO
                ),
            },
            {
                "id": "002",
                "file": None,  # No file for this row
                "video": ExternalAttachment(
                    url="https://vimeo.com/test", file_type=FileCellType.VIDEO
                ),
            },
        ],
    )

    # Mock S3 upload
    with patch.object(Attachment, "_upload_to_s3", return_value=True):
        dataset = datasets.create(dataset_request)

    # Verify both rows
    assert len(dataset.rows) == 2

    # First row should have both attachments
    row1 = dataset.rows[0]
    assert row1.values["file"]["storage"] == "internal"
    assert row1.values["video"]["storage"] == "external"
    assert row1.values["video"]["url"] == "https://youtube.com/watch?v=test"

    # Second row should have only video
    row2 = dataset.rows[1]
    assert row2.values["file"] is None
    assert row2.values["video"]["storage"] == "external"

    # Clean up
    os.unlink(test_file.name)


@pytest.mark.vcr
def test_create_dataset_with_in_memory_attachment(datasets):
    """Test creating dataset with Attachment from in-memory data."""

    # Create attachment from bytes
    image_data = b"fake image bytes"

    dataset_request = CreateDatasetRequest(
        slug="test-dataset-memory-attachment",
        name="Memory Attachment Dataset",
        columns=[
            ColumnDefinition(slug="name", name="Name", type=ColumnType.STRING),
            ColumnDefinition(slug="image", name="Image", type=ColumnType.FILE),
        ],
        rows=[
            {
                "name": "Test Item",
                "image": Attachment(
                    data=image_data,
                    filename="test.jpg",
                    content_type="image/jpeg",
                    file_type=FileCellType.IMAGE,
                ),
            }
        ],
    )

    # Mock S3 upload
    with patch.object(Attachment, "_upload_to_s3", return_value=True):
        dataset = datasets.create(dataset_request)

    # Verify
    assert len(dataset.rows) == 1
    row = dataset.rows[0]
    assert row.values["image"]["storage"] == "internal"
    assert row.values["image"]["type"] == "image"
    assert row.values["image"]["status"] == "success"


@pytest.mark.vcr
def test_create_dataset_without_attachments(datasets):
    """Test that create() method works normally for datasets without attachments."""

    dataset_request = CreateDatasetRequest(
        slug="test-dataset-no-attachments",
        name="Regular Dataset",
        columns=[
            ColumnDefinition(slug="id", name="ID", type=ColumnType.STRING),
            ColumnDefinition(slug="value", name="Value", type=ColumnType.NUMBER),
        ],
        rows=[
            {"id": "A", "value": 100},
            {"id": "B", "value": 200},
        ],
    )

    dataset = datasets.create(dataset_request)

    # Verify normal dataset creation
    assert dataset.slug == "test-dataset-no-attachments"
    assert len(dataset.rows) == 2
    assert dataset.rows[0].values["id"] == "A"
    assert dataset.rows[0].values["value"] == 100
    assert dataset.rows[1].values["id"] == "B"
    assert dataset.rows[1].values["value"] == 200



from typing import Dict, Any, Optional, Callable, TypeVar, ParamSpec, Union, Awaitable
from traceloop.sdk.evaluator.config import EvaluatorDetails
from traceloop.sdk.evaluator.evaluator import Evaluator
from .types import OutputSchema
import httpx
import asyncio
from functools import wraps


P = ParamSpec('P')
R = TypeVar('R')

# Type alias for evaluator specification - can be either a slug string or EvaluatorDetails
EvaluatorSpec = Union[str, EvaluatorDetails]


def guardrail(
    evaluator: EvaluatorSpec,
    on_evaluation_complete: Optional[Callable[[OutputSchema, Any], Any]] = None
) -> Callable[[Callable[P, Awaitable[Dict[str, Any]]]], Callable[P, Awaitable[Dict[str, Any]]]]:
    """
    Decorator that executes a guardrails evaluator on the decorated function's output.

    Args:
        evaluator: Either a slug string or an EvaluatorDetails object (with slug, version, config)
        on_evaluation_complete: Optional callback function that receives (evaluator_result, original_result)
                                and returns the final result. If not provided, returns original result on
                                success or an error message on failure.

    Returns:
        Result from on_evaluation_complete callback if provided, otherwise original result or error message
    """
    # Extract evaluator details as tuple (slug, version, config, required_fields) - same pattern as experiments
    slug: str
    evaluator_version: Optional[str]
    evaluator_config: Optional[Dict[str, Any]]
    required_input_fields: Optional[list[str]]

    if isinstance(evaluator, str):
        # Simple string slug - use default field mapping
        slug = evaluator
        evaluator_version = None
        evaluator_config = None
        required_input_fields = None
    elif isinstance(evaluator, EvaluatorDetails):
        # EvaluatorDetails object with config
        slug = evaluator.slug
        evaluator_version = evaluator.version
        evaluator_config = evaluator.config
        required_input_fields = evaluator.required_input_fields
    else:
        raise ValueError(f"evaluator must be str or EvaluatorDetails, got {type(evaluator)}")

    def decorator(func: Callable[P, R]) -> Callable[P, Dict[str, Any]]:
        @wraps(func)
        async def async_wrapper(*args: P.args, **kwargs: P.kwargs) -> Dict[str, Any]:
            # Execute the original function - should return a dict with fields matching required_input_fields
            result = await func(*args, **kwargs)  # type: ignore[misc]

            # Ensure we have a dict
            if not isinstance(result, dict):
                raise ValueError(
                    f"Function {func.__name__} must return a dict, got {type(result)}. "
                    f"Required fields: {required_input_fields or 'unknown'}"
                )

            original_result: Dict[str, Any] = result

            # Build evaluator_data based on required_input_fields or use all fields from result
            evaluator_data: Dict[str, str] = {}
            if required_input_fields:
                # Use only the required fields from the function result
                for field in required_input_fields:
                    if field not in original_result:
                        raise ValueError(
                            f"Function {func.__name__} must return dict with field '{field}'. "
                            f"Got: {list(original_result.keys())}"
                        )
                    evaluator_data[field] = str(original_result[field])
            else:
                # No required fields specified, use all fields from result
                for field, value in original_result.items():
                    evaluator_data[field] = str(value)

            try:
                from traceloop.sdk import Traceloop
                client_instance = Traceloop.get()
            except Exception as e:
                print(f"Warning: Could not get Traceloop client: {e}")
                return original_result

            evaluator_result = await client_instance.guardrails.execute_evaluator(
                slug, evaluator_data, evaluator_version, evaluator_config
            )

            # Use callback if provided, otherwise use default behavior
            if on_evaluation_complete:
                callback_result = on_evaluation_complete(evaluator_result, original_result)
                # Callback should return a dict, but we can't enforce this at compile time
                return callback_result  # type: ignore[no-any-return]
            else:
                # Default behavior: return original result (dict) regardless of pass/fail
                # Users should use callback for custom behavior
                return original_result

        @wraps(func)
        def sync_wrapper(*args: P.args, **kwargs: P.kwargs) -> Dict[str, Any]:

            # Execute the original function - should return a dict with fields matching required_input_fields
            result = func(*args, **kwargs)

            # Ensure we have a dict
            if not isinstance(result, dict):
                raise ValueError(
                    f"Function {func.__name__} must return a dict, got {type(result)}. "
                    f"Required fields: {required_input_fields or 'unknown'}"
                )

            original_result: Dict[str, Any] = result

            # Build evaluator_data based on required_input_fields or use all fields from result
            evaluator_data: Dict[str, str] = {}
            if required_input_fields:
                # Use only the required fields from the function result
                for field in required_input_fields:
                    if field not in original_result:
                        raise ValueError(
                            f"Function {func.__name__} must return dict with field '{field}'. "
                            f"Got: {list(original_result.keys())}"
                        )
                    evaluator_data[field] = str(original_result[field])
            else:
                # No required fields specified, use all fields from result
                for field, value in original_result.items():
                    evaluator_data[field] = str(value)

            # Get client instance
            try:
                from traceloop.sdk import Traceloop
                client_instance = Traceloop.get()
            except Exception as e:
                print(f"Warning: Could not get Traceloop client: {e}")
                return original_result

            loop = asyncio.get_event_loop()
            evaluator_result = loop.run_until_complete(
                client_instance.guardrails.execute_evaluator(
                    slug, evaluator_data, evaluator_version, evaluator_config
                )
            )

            # Use callback if provided, otherwise use default behavior
            if on_evaluation_complete:
                callback_result = on_evaluation_complete(evaluator_result, original_result)
                # Callback should return a dict, but we can't enforce this at compile time
                return callback_result  # type: ignore[no-any-return]
            else:
                # Default behavior: return original result (dict) regardless of pass/fail
                # Users should use callback for custom behavior
                return original_result

        # Return appropriate wrapper based on function type
        if asyncio.iscoroutinefunction(func):
            return async_wrapper  # type: ignore[return-value]
        else:
            return sync_wrapper

    return decorator  # type: ignore[return-value]


class Guardrails:
    """
    Guardrails class that wraps the Evaluator class for real-time evaluations.
    Unlike experiments, guardrails don't require task/experiment IDs.
    """

    def __init__(self, async_http_client: httpx.AsyncClient):
        self._evaluator = Evaluator(async_http_client)

    async def execute_evaluator(
        self,
        slug: str,
        data: Dict[str, str],
        evaluator_version: Optional[str] = None,
        evaluator_config: Optional[Dict[str, Any]] = None
    ) -> OutputSchema:
        """
        Execute evaluator for guardrails (real-time evaluation without experiment context).

        Args:
            slug: The evaluator slug to execute
            data: Input data mapping (guardrails format with InputExtractor)
            evaluator_version: Optional version of the evaluator
            evaluator_config: Optional configuration for the evaluator

        Returns:
            OutputSchema: The evaluation result with success/reason fields
        """
        try:

            # Use dummy IDs for guardrails (they don't need experiment tracking)
            result = await self._evaluator.run_experiment_evaluator(
                evaluator_slug=slug,
                task_id="guardrail",
                experiment_id="guardrail",
                experiment_run_id="guardrail",
                input=data,
                timeout_in_sec=120,
                evaluator_version=evaluator_version,
                evaluator_config=evaluator_config,
            )

            # Parse the result to OutputSchema format
            # Made by Traceloop evaluators return: {'evaluator_result': 'pass'/'fail', ...}
            raw_result = result.result

            # Check if this is a Made by Traceloop evaluator format
            if "evaluator_result" in raw_result:
                evaluator_result = raw_result.get("evaluator_result", "")
                success = evaluator_result == "pass"
                reason = raw_result.get("reason", None)
                return OutputSchema(reason=reason, success=success)

            # Otherwise, try custom evaluator format with 'pass' and 'reason' fields
            return OutputSchema.model_validate(raw_result)

        except Exception as e:
            print(f"Error executing evaluator {slug}. Error: {str(e)}")
            # Return a failure result on error
            return OutputSchema(reason=f"Evaluation error: {str(e)}", success=False)



import datetime
from enum import Enum
from typing import Any, Dict, List, Optional

from pydantic import BaseModel


class ColumnType(str, Enum):
    STRING = "string"
    NUMBER = "number"
    BOOLEAN = "boolean"
    JSON = "json"
    FILE = "file"


class FileCellType(str, Enum):
    IMAGE = "image"
    VIDEO = "video"
    AUDIO = "audio"
    FILE = "file"


class FileStorageType(str, Enum):
    INTERNAL = "internal"
    EXTERNAL = "external"


class FileCellMetadata(BaseModel):
    file_name: Optional[str] = None
    content_type: Optional[str] = None
    size_bytes: Optional[int] = None
    uploaded_at: Optional[datetime.datetime] = None
    thumbnail_url: Optional[str] = None
    thumbnail_key: Optional[str] = None


class FileCellValue(BaseModel):
    type: FileCellType
    status: str  # "in_progress", "success", "failed"
    storage: FileStorageType
    storage_key: Optional[str] = None
    url: Optional[str] = None
    metadata: Optional[FileCellMetadata] = None


# Internal models for API communication (not exposed directly to users)
class UploadURLRequest(BaseModel):
    type: FileCellType
    file_name: str
    content_type: Optional[str] = None
    with_thumbnail: bool = False
    metadata: Optional[Dict[str, Any]] = None


class UploadURLResponse(BaseModel):
    upload_url: str
    thumbnail_upload_url: Optional[str] = None
    storage_key: str
    thumbnail_key: Optional[str] = None
    expires_at: datetime.datetime
    method: str = "PUT"


class UploadStatusRequest(BaseModel):
    status: str  # "success", "failed"
    metadata: Optional[Dict[str, Any]] = None


class ExternalURLRequest(BaseModel):
    type: FileCellType
    url: str
    metadata: Optional[Dict[str, Any]] = None


class ColumnDefinition(BaseModel):
    slug: Optional[str] = None
    name: str
    type: ColumnType


ValuesMap = Dict[str, Any]


class CreateDatasetRequest(BaseModel):
    slug: str
    name: Optional[str] = None
    description: Optional[str] = None
    columns: Optional[List[ColumnDefinition]] = None
    rows: Optional[List[ValuesMap]] = None


class RowObject(BaseModel):
    id: str
    values: ValuesMap
    created_at: datetime.datetime
    updated_at: datetime.datetime


class CreateDatasetResponse(BaseModel):
    id: str
    slug: str
    name: str
    description: Optional[str] = None
    columns: Dict[str, ColumnDefinition]
    rows: Optional[List[RowObject]] = None
    last_version: Optional[str] = None
    created_at: datetime.datetime
    updated_at: datetime.datetime


class UpdateDatasetInput(BaseModel):
    name: Optional[str] = None
    description: Optional[str] = None


class CreateColumnInput(BaseModel):
    slug: str
    name: str
    type: ColumnType


class UpdateColumnInput(BaseModel):
    name: Optional[str] = None
    type: Optional[ColumnType] = None


class CreateRowsInput(BaseModel):
    rows: List[ValuesMap]


class CreateRowsResponse(BaseModel):
    rows: List[RowObject]
    total: int


class PublishDatasetResponse(BaseModel):
    dataset_id: str
    version: str


class AddColumnResponse(BaseModel):
    slug: str
    name: str
    type: ColumnType


class UpdateRowInput(BaseModel):
    values: ValuesMap


class DatasetMetadata(BaseModel):
    id: str
    slug: str
    name: str
    description: Optional[str] = None
    last_version: Optional[str] = None
    columns: Optional[Dict[str, ColumnDefinition]] = None
    created_at: datetime.datetime
    updated_at: datetime.datetime



import pytest
from traceloop.sdk import Traceloop, AssociationProperty
from traceloop.sdk.decorators import task, workflow
from opentelemetry.sdk.trace.export.in_memory_span_exporter import InMemorySpanExporter
from opentelemetry.semconv_ai import SpanAttributes


@pytest.fixture
def client_with_exporter():
    """
    Fixture that initializes Traceloop with API key.
    Client is only created when NO custom exporter/processor is provided.
    """
    # Initialize with API key and Traceloop endpoint - this creates a client
    client = Traceloop.init(
        app_name="test_associations",
        api_key="test-api-key",
        api_endpoint="https://api.traceloop.com",
        disable_batch=True,
        # NO exporter or processor - so client gets created
    )

    # Get spans from the tracer provider for assertions
    from opentelemetry import trace
    from opentelemetry.sdk.trace import TracerProvider

    tracer_provider = trace.get_tracer_provider()
    if isinstance(tracer_provider, TracerProvider):
        # Get the span processor's exporter
        span_processors = tracer_provider._active_span_processor._span_processors
        # Find the exporter from the processors
        for processor in span_processors:
            if hasattr(processor, 'span_exporter'):
                exporter = processor.span_exporter
                break
        else:
            # Fallback: create a mock exporter
            exporter = InMemorySpanExporter()
    else:
        exporter = InMemorySpanExporter()

    yield client, exporter

    # Cleanup
    if hasattr(exporter, 'clear'):
        exporter.clear()


def test_associations_create_single(client_with_exporter):
    """Test creating a single association."""
    client, exporter = client_with_exporter

    @workflow(name="test_single_association")
    def test_workflow():
        return test_task()

    @task(name="test_single_task")
    def test_task():
        return

    client.associations.set([(AssociationProperty.SESSION_ID, "conv-123")])
    test_workflow()

    spans = exporter.get_finished_spans()
    assert [span.name for span in spans] == [
        "test_single_task.task",
        "test_single_association.workflow",
    ]

    task_span = spans[0]
    workflow_span = spans[1]

    assert workflow_span.attributes[
        f"{SpanAttributes.TRACELOOP_ASSOCIATION_PROPERTIES}.session_id"
    ] == "conv-123"
    assert task_span.attributes[
        f"{SpanAttributes.TRACELOOP_ASSOCIATION_PROPERTIES}.session_id"
    ] == "conv-123"


def test_associations_create_multiple(client_with_exporter):
    """Test creating multiple associations at once."""
    client, exporter = client_with_exporter

    @workflow(name="test_multiple_associations")
    def test_workflow():
        return test_task()

    @task(name="test_multiple_task")
    def test_task():
        return

    client.associations.set([
        (AssociationProperty.USER_ID, "user-456"),
        (AssociationProperty.SESSION_ID, "session-789"),
        (AssociationProperty.CUSTOMER_ID, "customer-999"),
    ])
    test_workflow()

    spans = exporter.get_finished_spans()
    assert [span.name for span in spans] == [
        "test_multiple_task.task",
        "test_multiple_associations.workflow",
    ]

    task_span = spans[0]
    workflow_span = spans[1]

    # Check all associations are present
    assert workflow_span.attributes[
        f"{SpanAttributes.TRACELOOP_ASSOCIATION_PROPERTIES}.user_id"
    ] == "user-456"
    assert workflow_span.attributes[
        f"{SpanAttributes.TRACELOOP_ASSOCIATION_PROPERTIES}.session_id"
    ] == "session-789"
    assert workflow_span.attributes[
        f"{SpanAttributes.TRACELOOP_ASSOCIATION_PROPERTIES}.customer_id"
    ] == "customer-999"

    assert task_span.attributes[
        f"{SpanAttributes.TRACELOOP_ASSOCIATION_PROPERTIES}.user_id"
    ] == "user-456"
    assert task_span.attributes[
        f"{SpanAttributes.TRACELOOP_ASSOCIATION_PROPERTIES}.session_id"
    ] == "session-789"
    assert task_span.attributes[
        f"{SpanAttributes.TRACELOOP_ASSOCIATION_PROPERTIES}.customer_id"
    ] == "customer-999"


def test_associations_within_workflow(client_with_exporter):
    """Test creating associations within a workflow."""
    client, exporter = client_with_exporter

    @workflow(name="test_associations_within")
    def test_workflow():
        client.associations.set([
            (AssociationProperty.SESSION_ID, "conv-abc"),
            (AssociationProperty.USER_ID, "user-xyz"),
        ])
        return test_task()

    @task(name="test_within_task")
    def test_task():
        return

    test_workflow()

    spans = exporter.get_finished_spans()
    assert [span.name for span in spans] == [
        "test_within_task.task",
        "test_associations_within.workflow",
    ]

    task_span = spans[0]
    workflow_span = spans[1]

    # Both spans should have all associations
    assert workflow_span.attributes[
        f"{SpanAttributes.TRACELOOP_ASSOCIATION_PROPERTIES}.session_id"
    ] == "conv-abc"
    assert workflow_span.attributes[
        f"{SpanAttributes.TRACELOOP_ASSOCIATION_PROPERTIES}.user_id"
    ] == "user-xyz"

    assert task_span.attributes[
        f"{SpanAttributes.TRACELOOP_ASSOCIATION_PROPERTIES}.session_id"
    ] == "conv-abc"
    assert task_span.attributes[
        f"{SpanAttributes.TRACELOOP_ASSOCIATION_PROPERTIES}.user_id"
    ] == "user-xyz"


def test_all_association_properties(client_with_exporter):
    """Test that all AssociationProperty enum values work correctly."""
    client, exporter = client_with_exporter

    @workflow(name="test_all_properties")
    def test_workflow():
        return

    client.associations.set([
        (AssociationProperty.SESSION_ID, "conv-1"),
        (AssociationProperty.CUSTOMER_ID, "customer-2"),
        (AssociationProperty.USER_ID, "user-3"),
        (AssociationProperty.SESSION_ID, "session-4"),
    ])
    test_workflow()

    spans = exporter.get_finished_spans()
    workflow_span = spans[0]

    assert workflow_span.attributes[
        f"{SpanAttributes.TRACELOOP_ASSOCIATION_PROPERTIES}.customer_id"
    ] == "customer-2"
    assert workflow_span.attributes[
        f"{SpanAttributes.TRACELOOP_ASSOCIATION_PROPERTIES}.user_id"
    ] == "user-3"
    assert workflow_span.attributes[
        f"{SpanAttributes.TRACELOOP_ASSOCIATION_PROPERTIES}.session_id"
    ] == "session-4"



from traceloop.sdk.tracing.context_manager import get_tracer as get_tracer
from traceloop.sdk.tracing.tracing import (
    set_workflow_name as set_workflow_name,
    set_agent_name as set_agent_name,
    set_conversation_id as set_conversation_id,
)



"""
Attachment classes for handling file uploads and downloads in datasets.
Simplified implementation inspired by Braintrust's attachment pattern.
"""

import mimetypes
import os
from pathlib import Path
from typing import Any, Dict, Optional

import requests
from traceloop.sdk.client.http import HTTPClient

from .model import (
    ExternalURLRequest,
    FileCellType,
    FileStorageType,
    UploadStatusRequest,
    UploadURLRequest,
    UploadURLResponse,
)


class Attachment:
    """
    Represents a file to be uploaded to a dataset cell.
    Supports both file paths and in-memory data.
    """

    def __init__(
        self,
        file_path: Optional[str] = None,
        data: Optional[bytes] = None,
        filename: Optional[str] = None,
        content_type: Optional[str] = None,
        file_type: Optional[FileCellType] = None,
        metadata: Optional[Dict[str, Any]] = None,
        with_thumbnail: bool = False,
        thumbnail_path: Optional[str] = None,
        thumbnail_data: Optional[bytes] = None,
    ):
        # Validate input
        if file_path and data:
            raise ValueError("Cannot provide both file_path and data")
        if not file_path and not data:
            raise ValueError("Must provide either file_path or data")

        self.file_path = file_path
        self.data = data
        self.metadata = metadata or {}
        self.with_thumbnail = with_thumbnail
        self.thumbnail_path = thumbnail_path
        self.thumbnail_data = thumbnail_data

        # Set filename
        if filename:
            self.filename = filename
        elif file_path:
            self.filename = os.path.basename(file_path)
        else:
            self.filename = "attachment"

        # Set content type
        if content_type:
            self.content_type = content_type
        elif file_path:
            self.content_type = (
                mimetypes.guess_type(file_path)[0] or "application/octet-stream"
            )
        else:
            self.content_type = "application/octet-stream"

        # Set file type
        if file_type:
            self.file_type = file_type
        else:
            self.file_type = self._guess_file_type()

    def _guess_file_type(self) -> FileCellType:
        """Guess file type from content type."""
        if self.content_type.startswith("image/"):
            return FileCellType.IMAGE
        elif self.content_type.startswith("video/"):
            return FileCellType.VIDEO
        elif self.content_type.startswith("audio/"):
            return FileCellType.AUDIO
        else:
            return FileCellType.FILE

    def _get_file_data(self) -> bytes:
        """Get file data as bytes."""
        if self.data is not None:
            return self.data
        elif self.file_path:
            if not os.path.exists(self.file_path):
                raise FileNotFoundError(f"File not found: {self.file_path}")
            with open(self.file_path, "rb") as f:
                return f.read()
        raise ValueError("No file data available")

    def _get_file_size(self) -> int:
        """Get file size in bytes."""
        if self.data is not None:
            return len(self.data)
        elif self.file_path and os.path.exists(self.file_path):
            return os.path.getsize(self.file_path)
        return 0

    def upload(
        self,
        http_client: HTTPClient,
        dataset_slug: str,
        row_id: str,
        column_slug: str,
    ) -> "AttachmentReference":
        """Upload the attachment to a dataset cell."""
        # Request upload URL
        request = UploadURLRequest(
            type=self.file_type,
            file_name=self.filename,
            content_type=self.content_type,
            with_thumbnail=self.with_thumbnail,
            metadata=self.metadata,
        )

        result = http_client.post(
            f"datasets/{dataset_slug}/rows/{row_id}/cells/{column_slug}/upload-url",
            request.model_dump(),
        )

        if not result:
            raise Exception(f"Failed to get upload URL for {column_slug}")

        upload_response = UploadURLResponse(**result)

        # Upload to S3
        if not self._upload_to_s3(upload_response.upload_url):
            raise Exception(f"Failed to upload {self.filename}")

        # Upload thumbnail if provided
        if self.with_thumbnail and upload_response.thumbnail_upload_url:
            if self.thumbnail_data is not None:
                thumb_bytes = self.thumbnail_data
            elif self.thumbnail_path:
                with open(self.thumbnail_path, "rb") as f:
                    thumb_bytes = f.read()
            else:
                thumb_bytes = None
            if thumb_bytes is not None:
                requests.put(upload_response.thumbnail_upload_url, data=thumb_bytes)

        # Confirm upload
        metadata = self.metadata.copy()
        metadata["size_bytes"] = self._get_file_size()

        status_request = UploadStatusRequest(status="success", metadata=metadata)
        http_client.put(
            f"datasets/{dataset_slug}/rows/{row_id}/cells/{column_slug}/upload-status",
            status_request.model_dump(),
        )

        return AttachmentReference(
            storage_type=FileStorageType.INTERNAL,
            storage_key=upload_response.storage_key,
            file_type=self.file_type,
            metadata=metadata,
        )

    def _upload_to_s3(self, upload_url: str) -> bool:
        """Upload file to S3."""
        try:
            file_data = self._get_file_data()
            response = requests.put(
                upload_url, data=file_data, headers={"Content-Type": self.content_type}
            )
            return response.status_code in [200, 201, 204]
        except Exception:
            return False


class ExternalAttachment:
    """
    Represents an external file URL to be linked to a dataset cell.
    """

    def __init__(
        self,
        url: str,
        filename: Optional[str] = None,
        content_type: Optional[str] = None,
        file_type: FileCellType = FileCellType.FILE,
        metadata: Optional[Dict[str, Any]] = None,
    ):
        self.url = url
        self.filename = filename or url.split("/")[-1]
        self.content_type = content_type
        self.file_type = file_type
        self.metadata = metadata or {}

    def attach(
        self,
        http_client: HTTPClient,
        dataset_slug: str,
        row_id: str,
        column_slug: str,
    ) -> "AttachmentReference":
        """Attach external URL to a dataset cell."""
        request = ExternalURLRequest(
            type=self.file_type,
            url=self.url,
            metadata=self.metadata,
        )

        result = http_client.post(
            f"datasets/{dataset_slug}/rows/{row_id}/cells/{column_slug}/external-url",
            request.model_dump(),
        )

        if not result:
            raise Exception(f"Failed to set external URL for {column_slug}")

        return AttachmentReference(
            storage_type=FileStorageType.EXTERNAL,
            url=self.url,
            file_type=self.file_type,
            metadata=self.metadata,
        )


class AttachmentReference:
    """
    Reference to an attachment in a dataset cell.
    """

    def __init__(
        self,
        storage_type: FileStorageType,
        storage_key: Optional[str] = None,
        url: Optional[str] = None,
        file_type: Optional[FileCellType] = None,
        metadata: Optional[Dict[str, Any]] = None,
        http_client: Optional[HTTPClient] = None,
        dataset_slug: Optional[str] = None,
    ):
        self.storage_type = storage_type
        self.storage_key = storage_key
        self.url = url
        self.file_type = file_type
        self.metadata = metadata or {}
        self.http_client = http_client
        self.dataset_slug = dataset_slug
        self._cached_data: Optional[bytes] = None

    @property
    def data(self) -> bytes:
        """Download and return attachment data as bytes."""
        if self._cached_data is None:
            download_url = self.get_url()
            if not download_url:
                raise Exception("No download URL available")
            response = requests.get(download_url)
            response.raise_for_status()
            self._cached_data = response.content
        return self._cached_data

    def download(self, file_path: Optional[str] = None) -> Optional[bytes]:
        """Download the attachment."""
        file_data = self.data
        if file_path:
            Path(file_path).parent.mkdir(parents=True, exist_ok=True)
            with open(file_path, "wb") as f:
                f.write(file_data)
            return None
        return file_data

    def get_url(self) -> Optional[str]:
        """Get download URL for the attachment."""
        if self.storage_type == FileStorageType.EXTERNAL:
            return self.url
        # For internal storage, would need to implement presigned URL generation
        return None

    def __repr__(self) -> str:
        """String representation."""
        if self.storage_type == FileStorageType.EXTERNAL:
            return f"<AttachmentReference(external, url={self.url})>"
        else:
            return f"<AttachmentReference(internal, key={self.storage_key})>"



from traceloop.sdk.datasets.attachment import (
    Attachment,
    AttachmentReference,
    ExternalAttachment,
)
from traceloop.sdk.datasets.base import BaseDatasetEntity
from traceloop.sdk.datasets.column import Column
from traceloop.sdk.datasets.dataset import Dataset
from traceloop.sdk.datasets.model import (
    ColumnType,
    DatasetMetadata,
    FileCellType,
    FileStorageType,
)
from traceloop.sdk.datasets.row import Row

__all__ = [
    "Dataset",
    "Column",
    "Row",
    "BaseDatasetEntity",
    "ColumnType",
    "DatasetMetadata",
    "FileCellType",
    "FileStorageType",
    "Attachment",
    "ExternalAttachment",
    "AttachmentReference",
]
